/* tok.c by Alain Brobecker */
/* 1999-10-12 First working version, without 2-bytes tokens (CASE...)   */
/* 1999-10-13 Changed ELSE to &cc, now correctly tokenise permute.bas   */
/* 1999-10-15 Alfa sorted tokens, checked that ENDPROC is before END... */
/* Exchange "_write" and "write" between dos/unix versions */

/*Format of a BBC basic file (filetype &ffb)
  0D    - new line indicator
  00 01 - line number
  2F    - size of line (with the 0D)
  Here is the tokenised line (normally (2F)-4 bytes)
  0D FF - end of file
*/

/*!!! ELSE=&cc if it's the first instruction of a line? */
/*!!! ELSE=&8b if it's on the same line as THEN? */

#include <stdio.h>

/***************************************************************************/
/* C SUCKS! BASIC RULES! Luckily C has the #define metainstruction.        */
/***************************************************************************/
#define AND             &           /*****Logical operators for expressions*/
#define OR              |
#define EOR             ^
#define NOT             ~
#define FALSE           0
#define TRUE            -1
#define and             &&          /******Logical operators for conditions*/
#define or              ||
#define not             !

#define IF(expr)        if(expr){   /*************Flow control instructions*/
#define ELSE            } else {
#define ENDIF           }
#define FOR(expr0,expr1,expr2) for(expr0;expr1;expr2){
#define NEXT            }
#define REPEAT          do {
#define UNTIL(expr)     } while(!expr);
#define WHILE(expr)     while(expr){
#define ENDWHILE        }
#define CASE(expression) switch(expression){
#define WHEN            case
#define ENDCASE         }

/**** Global program vars **************************************************/
unsigned int LineNb;
char *PtrIn,*PtrOut,*PtrSize,*PtrEnd,*PtrTok;
char In[29999],Out[29999];

/**** Tokens table *********************************************************/
/* Unimplemented tokens: &8d; &c6; &c7; &c8 (3 lasts for 2-bytes tokens).  */
/* Some tokens are repeated twice(!?): ELSE, PTR, PAGE, TIME, LOMEM, HIMEM */
/* For every start letter                                                  */
/*   1 byte = N = length-1 of keyword in ascii, or 0 to terminate          */
/*   N bytes = keyword excluding first letter (eg ND for AND...).          */
/*   1 byte = single byte token associated to keyword                      */
char *token[26] = {
/*A*/ "\002BS\224\002CS\225\004DVAL\226\002ND\200\002SC\227\002SN\230\002TN\231",
/*B*/ "\003GET\232\003PUT\325",
/*C*/ "\003ALL\326\003HR$\275\004HAIN\327\004LEAR\330\002LG\332\004LOSE\331\002LS\333\005OLOUR\373\002OS\233\004OUNT\234",
/*D*/ "\003ATA\334\002EF\335\002EG\235\002IM\336\002IV\201\003RAW\337",
/*E*/ "\003LSE\314\003LSE\213\006NDCASE\313\004NDIF\315\006NDPROC\341\007NDWHILE\316\002ND\340\007NVELOPE\342\002OF\305\002OR\202\002RL\236\004RROR\205\002RR\237\003VAL\240\002XP\241\002XT\242",
/*F*/ "\004ALSE\243\001N\244\002OR\343",
/*G*/ "\003COL\346\003ET$\276\002ET\245\004OSUB\344\003OTO\345",
/*H*/ "\004IMEM\223\004IMEM\323",
/*I*/ "\001F\347\005NKEY$\277\004NKEY\246\004NPUT\350\005NSTR(\247\002NT\250",
/*J*/ "",/*K*/ "",
/*L*/ "\005EFT$(\300\002ET\351\002EN\251\003INE\206\001N\252\004OCAL\352\002OG\253\004OMEM\222\004OMEM\322",
/*M*/ "\004ID$(\301\003ODE\353\002OD\203\003OVE\354",
/*N*/ "\003EXT\355\002OT\254",
/*O*/ "\002FF\207\001F\312\001N\356\005PENIN\216\006PENOUT\256\005PENUP\255\001R\204\004SCLI\377\010THERWISE\177",
/*P*/ "\003AGE\220\003AGE\320\001I\257\003LOT\360\005OINT(\260\002OS\261\004RINT\361\003ROC\362\002TR\217\002TR\317",
/*Q*/ "",
/*R*/ "\002AD\262\003EAD\363\002EM\364\005EPEAT\365\005EPORT\366\006ESTORE\367\005ETURN\370\006IGHT$(\302\002ND\263\002UN\371",
/*S*/ "\002IN\265\002GN\264\004OUND\325\002PC\211\002QR\266\003TEP\210\003TOP\372\003TR$\303\007TRING$(\304",
/*T*/ "\003AB(\212\002AN\267\003HEN\214\003IME\221\003IME\321\001O\270\004RACE\374\003RUE\271",
/*U*/ "\004NTIL\375\002SR\272",
/*V*/ "\002AL\273\002DU\357\003POS\274",
/*W*/ "\003HEN\311\004IDTH\376",
/*X*/ "",/*Y*/ "",/*Z*/ ""
};

/**** Newline **************************************************************/
/* Save size of finished line, and put bytes+LineNb for new one. */
void NewLine() {
  IF(PtrOut-PtrSize+3>255) printf("!!!Error: Line %d exceeding 255 bytes.\n",LineNb); ENDIF
  IF(LineNb>32767) printf("!!!Error: Too many lines.\n"); ENDIF
  *PtrSize=PtrOut-PtrSize+3;    /*write length of previous line*/
  PtrSize=PtrOut+3;             /*pointer on length for current line*/
  *PtrOut++=0x0d;
  *PtrOut++=LineNb>>8;          /*write LineNb*/
  *PtrOut=LineNb AND 255;
  PtrOut=PtrOut+2;              /*skip*/
  LineNb+=1;
}

/**** Compare Len chars of two strings *************************************/
/* And returns FALSE or TRUE accordingly. */
int Compare(char *Str1,char *Str2,char Len) {
  FOR(,Len>0,Len--)
    IF(*Str1++!=*Str2++) return FALSE; ENDIF
  NEXT
  return TRUE;
}

/**** Tok ******************************************************************/
int main() {
int Handle,Found;
char c,LenTok;
unsigned short int toto;
  printf("Usage: tok <InFile\nReturns tokenised file as 'ffb'\n");
/*** get input file ***/
  PtrEnd=&In[0];
  WHILE(!feof(stdin))
    *PtrEnd++=getc(stdin);
  ENDWHILE
  printf("In: %d bytes\n",PtrEnd-&In[0]);
/*** tokenise ***/
  PtrIn=&In[0];
  PtrOut=&Out[0];
  PtrSize=PtrOut+3;             /*Fake PtrSize for first call to NewLine*/
  LineNb=1;
  WHILE(PtrIn<PtrEnd)
    NewLine();
    c=*PtrIn++;
    WHILE(c!=0x0a && PtrIn<PtrEnd)
      CASE(c)
        WHEN 0x0d: break;       /*just skip*/
        WHEN '"':               /*copy until next " (inclusive)*/
          *PtrOut++=c; c=*PtrIn++;
          WHILE(c!='"') *PtrOut++=c; c=*PtrIn++; ENDWHILE
          *PtrOut++=c;
          break;
        default:
          IF(c>='A' && c<='Z') /*uppercase letter?*/
            PtrTok=token[c-65];
            LenTok=*PtrTok++;
            Found=FALSE;
            WHILE(Found==FALSE && LenTok!=0) /*compare with keywords*/
              Found=Compare(PtrIn,PtrTok,LenTok);
              PtrTok=PtrTok+LenTok+1;
              IF(Found==FALSE)
                LenTok=*PtrTok++;
              ENDIF
            ENDWHILE
            IF(Found==TRUE)     /*keyword found, insert according token*/
              PtrIn=PtrIn+LenTok;
              *PtrOut++=*--PtrTok;
            ELSE
              *PtrOut++=c;
            ENDIF
          ELSE
            *PtrOut++=c;
          ENDIF
      ENDCASE
      c=*PtrIn++;
    ENDWHILE
  ENDWHILE
  *PtrSize=PtrOut-PtrSize+3;    /*write length of last line*/
  *PtrOut++=0x0d;         /*basic file terminator*/
  *PtrOut++=0xff;
/*** save output file ***/
  Handle=creat("ffb",0775);
  IF(Handle==FALSE) printf("!!!Error: while saving\n"); return 1; ENDIF /*problem?*/
  _write(Handle,&Out[0],PtrOut-&Out[0]);
  close(Handle);
  printf("Out: %d bytes\n",PtrOut-&Out[0]);
}
